@@ -22,15 +22,6 @@ __all__ = ['BlockCode', 'Heading', 'Quote', 'CodeFence', 'ThematicBreak',
            'List', 'Table', 'Footnote', 'Paragraph']
 
 
-"""
-Stores a reference to the current document token.
-
-When parsing, footnote entries will be stored in the document by
-accessing this pointer.
-"""
-_root_node = None
-
-
 def tokenize(lines):
     """
     A wrapper around block_tokenizer.tokenize. Pass in all block-level
@@ -143,12 +134,9 @@ class Document(BlockToken):
             lines = lines.splitlines(keepends=True)
         lines = [line if line.endswith('\n') else '{}\n'.format(line) for line in lines]
         self.footnotes = {}
-        global _root_node
-        _root_node = self
-        span_token._root_node = self
+        token._root_node = self
         self.children = tokenize(lines)
-        span_token._root_node = None
-        _root_node = None
+        token._root_node = None
 
 
 class Heading(BlockToken):
@@ -324,14 +312,13 @@ class Paragraph(BlockToken):
                 and not Quote.start(next_line)):
 
             # check if next_line starts List
-            list_pair = ListItem.parse_marker(next_line)
-            if (len(next_line) - len(next_line.lstrip()) < 4
-                    and list_pair is not None):
-                prepend, leader = list_pair
-                # non-empty list item
-                if next_line[:prepend].endswith(' '):
-                    # unordered list, or ordered list starting from 1
-                    if not leader[:-1].isdigit() or leader[:-1] == '1':
+            marker_tuple = ListItem.parse_marker(next_line)
+            if (marker_tuple is not None):
+                _, leader, content = marker_tuple
+                # to break a paragraph, the first line may not be empty,
+                # and the list must be unordered or start from 1.
+                if not content.strip() == '':
+                    if not leader[0].isdigit() or leader in ['1.', '1)']:
                         break
 
             # check if next_line starts HTMLBlock other than type 7
@@ -427,7 +414,9 @@ class CodeFence(BlockToken):
         if not match_obj:
             return False
         prepend, leader, lang = match_obj.groups()
-        if leader[0] in lang or leader[0] in line[match_obj.end():]:
+        # info strings for backtick code blocks may not contain backticks,
+        # but info strings for tilde code blocks may contain both tildes and backticks.
+        if leader[0] == '`' and '`' in line[match_obj.end(2):]:
             return False
         cls._open_info = len(prepend), leader, lang
         return True
@@ -512,7 +501,8 @@ class ListItem(BlockToken):
     Not included in the parsing process, but called by List.
     """
     repr_attributes = ("leader", "prepend", "loose")
-    pattern = re.compile(r'\s*(\d{0,9}[.)]|[+\-*])(\s*$|\s+)')
+    pattern = re.compile(r' {0,3}(\d{0,9}[.)]|[+\-*])($|\s+)')
+    continuation_pattern = re.compile(r'([ \t]*)(\S.*\n|\n)')
 
     def __init__(self, parse_buffer, prepend, leader):
         self.leader = leader
@@ -520,9 +510,23 @@ class ListItem(BlockToken):
         self.children = tokenizer.make_tokens(parse_buffer)
         self.loose = parse_buffer.loose
 
-    @staticmethod
-    def in_continuation(line, prepend):
-        return line.strip() == '' or len(line) - len(line.lstrip()) >= prepend
+    @classmethod
+    def parse_continuation(cls, line, prepend):
+        """
+        Returns content (i.e. the line with the prepend stripped off) iff the line
+        is a valid continuation line for a list item with the given prepend length,
+        otherwise None.
+
+        Note that the list item may still continue even if this test doesn't pass
+        due to lazy continuation.
+        """
+        match_obj = cls.continuation_pattern.match(line)
+        if match_obj is None:
+            return None
+        if match_obj.group(2) == '\n':
+            return '\n'
+        expanded_spaces = match_obj.group(1).expandtabs(4)
+        return expanded_spaces[prepend:] + match_obj.group(2) if len(expanded_spaces) >= prepend else None
 
     @staticmethod
     def other_token(line):
@@ -534,89 +538,93 @@ class ListItem(BlockToken):
     @classmethod
     def parse_marker(cls, line):
         """
-        Returns a pair (prepend, leader) iff the line has a valid leader.
+        Returns a tuple (prepend, leader, content) iff the line has a valid leader and at
+        least one space separating leader and content, or if the content is empty, in which
+        case there need not be any spaces.
+        The return value is None if the line doesn't have a valid marker.
+
+        The leader is a bullet list marker, or an ordered list marker.
+
+        The prepend is the start position of the content, i.e., the indentation required
+        for continuation lines.
         """
         match_obj = cls.pattern.match(line)
         if match_obj is None:
-            return None        # no valid leader
+            return None
+        prepend = len(match_obj.group(0).expandtabs(4))
         leader = match_obj.group(1)
-        content = match_obj.group(0).replace(leader+'\t', leader+'   ', 1)
-        # reassign prepend and leader
-        prepend = len(content)
-        if prepend == len(line.rstrip('\n')):
-            prepend = match_obj.end(1) + 1
-        else:
-            spaces = match_obj.group(2)
-            if spaces.startswith('\t'):
-                spaces = spaces.replace('\t', '   ', 1)
-            spaces = spaces.replace('\t', '    ')
-            n_spaces = len(spaces)
-            if n_spaces > 4:
-                prepend = match_obj.end(1) + 1
-        return prepend, leader
+        content = line[match_obj.end(0):]
+        n_spaces = prepend - match_obj.end(1)
+        if n_spaces > 4:
+            # if there are more than 4 spaces after the leader, we treat them as part of the content
+            # with the exception of the first (marker separator) space.
+            prepend -= n_spaces - 1
+            content = ' ' * (n_spaces - 1) + content
+        return prepend, leader, content
 
     @classmethod
     def read(cls, lines, prev_marker=None):
         next_marker = None
         lines.anchor()
-        prepend = -1
-        leader = None
         line_buffer = []
 
         # first line
         line = next(lines)
-        prepend, leader = prev_marker if prev_marker else cls.parse_marker(line)
-        line = line.replace(leader+'\t', leader+'   ', 1).replace('\t', '    ')
-        empty_first_line = line[prepend:].strip() == ''
-        if not empty_first_line:
-            line_buffer.append(line[prepend:])
         next_line = lines.peek()
-        if empty_first_line and next_line is not None and next_line.strip() == '':
-            parse_buffer = tokenizer.tokenize_block([next(lines)], _token_types)
-            next_line = lines.peek()
-            if next_line is not None:
-                marker_info = cls.parse_marker(next_line)
-                if marker_info is not None:
-                    next_marker = marker_info
-            return (parse_buffer, prepend, leader), next_marker
+        prepend, leader, content = prev_marker if prev_marker else cls.parse_marker(line)
+        if content.strip() == '':
+            # item starting with a blank line: look for the next non-blank line
+            prepend = len(leader) + 1
+            blanks = 1
+            while next_line is not None and next_line.strip() == '':
+                blanks += 1
+                next(lines)
+                next_line = lines.peek()
+            # if the line following the list marker is also empty, then this is an empty
+            # list item.
+            if blanks > 1:
+                parse_buffer = tokenizer.ParseBuffer()
+                parse_buffer.loose = True
+                next_marker = cls.parse_marker(next_line) if next_line is not None else None
+                return (parse_buffer, prepend, leader), next_marker
+        else:
+            line_buffer.append(content)
 
-        # loop
-        newline = 0
+        # loop over the following lines, looking for the end of the list item
+        newline_count = 0
         while True:
-            # no more lines
             if next_line is None:
-                # strip off newlines
-                if newline:
+                # list item ends here because we have reached the end of content
+                if newline_count:
                     lines.backstep()
-                    del line_buffer[-newline:]
+                    del line_buffer[-newline_count:]
                 break
-            next_line = next_line.replace('\t', '    ')
-            # not in continuation
-            if not cls.in_continuation(next_line, prepend):
-                # directly followed by another token
+
+            continuation = cls.parse_continuation(next_line, prepend)
+            if not continuation:
+                # the line doesn't have the indentation to show that it belongs to
+                # the list item, but it should be included anyway by lazy continuation...
+                # ...unless it's the start of another token
                 if cls.other_token(next_line):
-                    if newline:
+                    if newline_count:
                         lines.backstep()
-                        del line_buffer[-newline:]
+                        del line_buffer[-newline_count:]
                     break
-                # next_line is a new list item
+                # ...or it's a new list item
                 marker_info = cls.parse_marker(next_line)
                 if marker_info is not None:
                     next_marker = marker_info
                     break
-                # not another item, has newlines -> not continuation
-                if newline:
+                # ...or the line above it was blank
+                if newline_count:
                     lines.backstep()
-                    del line_buffer[-newline:]
+                    del line_buffer[-newline_count:]
                     break
+                continuation = next_line
+
+            line_buffer.append(continuation)
+            newline_count = newline_count + 1 if continuation == '\n' else 0
             next(lines)
-            line = next_line
-            stripped = line.lstrip(' ')
-            diff = len(line) - len(stripped)
-            if diff > prepend:
-                stripped = ' ' * (diff - prepend) + stripped
-            line_buffer.append(stripped)
-            newline = newline + 1 if next_line.strip() == '' else 0
             next_line = lines.peek()
 
         # block-level tokens are parsed here, so that footnotes can be
@@ -732,10 +740,6 @@ class Footnote(BlockToken):
     The constructor returns None, because the footnote information
     is stored in Footnote.read.
     """
-    # Not used, matched manually instead.
-    # We also rely on code block and similar being parsed beforehand here.
-    label_pattern = re.compile(r'[ \n]{0,3}\[(.+?)\]', re.DOTALL)
-
     def __new__(cls, _):
         return None
 
@@ -754,74 +758,114 @@ class Footnote(BlockToken):
         offset = 0
         matches = []
         while offset < len(string) - 1:
-            match_info = cls.match_reference(lines, string, offset)
+            match_info = cls.match_reference(string, offset)
             if match_info is None:
+                # backtrack the lines that have not been consumed
+                lines._index -= string[offset:].count('\n')
                 break
             offset, match = match_info
             matches.append(match)
-        cls.append_footnotes(matches, _root_node)
+        cls.append_footnotes(matches, token._root_node)
         return matches or None
 
     @classmethod
-    def match_reference(cls, lines, string, offset):
+    def match_reference(cls, string, offset):
+        # up to three spaces, "[", label, "]"
         match_info = cls.match_link_label(string, offset)
         if not match_info:
-            cls.backtrack(lines, string, offset)
             return None
         _, label_end, label = match_info
 
+        # ":"
         if not follows(string, label_end-1, ':'):
-            cls.backtrack(lines, string, offset)
             return None
 
-        match_info = cls.match_link_dest(string, label_end)
+        # optional spaces or tabs (including up to one line ending)
+        dest_start = shift_whitespace(string, label_end + 1)
+        if dest_start == len(string):
+            return None
+
+        # link destination
+        match_info = cls.match_link_dest(string, dest_start)
         if not match_info:
-            cls.backtrack(lines, string, offset)
             return None
         _, dest_end, dest = match_info
 
-        match_info = cls.match_link_title(string, dest_end)
-        if not match_info:
-            cls.backtrack(lines, string, dest_end)
+        # either of:
+        # 1) optional spaces or tabs and then a line break to finish the link reference definition.
+        # 2) optional spaces or tabs (including up to one line ending) followed by a title.
+        # in any case, if the destination is followed directly by non-whitespace, then it's not
+        # a valid link reference definition.
+        title_start = shift_whitespace(string, dest_end)
+        if title_start == dest_end and title_start < len(string):
             return None
+
+        # link title
+        match_info = cls.match_link_title(string, title_start)
+        if not match_info:
+            # no valid title found. if there was a line break following the destination,
+            # we still have a valid link reference definition. otherwise not.
+            eol_pos = string[dest_end:title_start].find("\n")
+            if eol_pos >= 0:
+                return dest_end + eol_pos + 1, (label, dest, "")
+            else:
+                return None
         _, title_end, title = match_info
 
-        return title_end, (label, dest, title)
+        # optional spaces or tabs. final line ending.
+        line_end = title_end
+        while line_end < len(string):
+            if string[line_end] == '\n':
+                return line_end + 1, (label, dest, title)
+            elif string[line_end] in whitespace:
+                line_end += 1
+            else:
+                break
+
+        # non-whitespace found on the same line as the title, making it invalid.
+        # if there was a line break following the destination,
+        # we still have a valid link reference definition. otherwise not.
+        eol_pos = string[dest_end:title_start].find("\n")
+        if eol_pos >= 0:
+            return dest_end + eol_pos + 1, (label, dest, "")
+        else:
+            return None
 
     @classmethod
     def match_link_label(cls, string, offset):
+        """
+        Matches: up to three spaces, "[", label, "]".
+        """
         start = -1
-        end = -1
         escaped = False
         for i, c in enumerate(string[offset:], start=offset):
-            if c == '\\' and not escaped:
+            if escaped:
+                escaped = False
+            elif c == '\\':
                 escaped = True
-            elif c == '[' and not escaped:
+            elif c == '[':
                 if start == -1:
                     start = i
                 else:
                     return None
-            elif c == ']' and not escaped:
-                end = i
-                label = string[start+1:end]
+            elif c == ']':
+                label = string[start+1:i]
                 if label.strip() != '':
-                    return start, end+1, label
+                    return start, i+1, label
+                return None
+            # only spaces allowed before the opening bracket
+            if start == -1 and not (c == " " and i - offset < 3):
                 return None
-            elif escaped:
-                escaped = False
         return None
 
     @classmethod
     def match_link_dest(cls, string, offset):
-        offset = shift_whitespace(string, offset+1)
-        if offset == len(string):
-            return None
         if string[offset] == '<':
             escaped = False
             for i, c in enumerate(string[offset+1:], start=offset+1):
                 if c == '\\' and not escaped:
                     escaped = True
-                elif c == ' ' or c == '\n' or (c == '<' and not escaped):
+                elif c == '\n' or (c == '<' and not escaped):
                     return None
                 elif c == '>' and not escaped:
                     return offset, i+1, string[offset+1:i]
@@ -851,32 +895,22 @@ class Footnote(BlockToken):
 
     @classmethod
     def match_link_title(cls, string, offset):
-        new_offset = shift_whitespace(string, offset)
-        if (new_offset == len(string)
-                or '\n' in string[offset:new_offset]
-                and string[new_offset] == '['):
-            return offset, new_offset, ''
-        if string[new_offset] == '"':
+        if offset == len(string):
+            return None
+        if string[offset] == '"':
             closing = '"'
-        elif string[new_offset] == "'":
+        elif string[offset] == "'":
             closing = "'"
-        elif string[new_offset] == '(':
+        elif string[offset] == '(':
             closing = ')'
-        elif '\n' in string[offset:new_offset]:
-            return offset, offset, ''
         else:
-            # XXX: Can this actually ever happen?
             return None
-        offset = new_offset
         escaped = False
         for i, c in enumerate(string[offset+1:], start=offset+1):
             if c == '\\' and not escaped:
                 escaped = True
             elif c == closing and not escaped:
-                new_offset = shift_whitespace(string, i+1)
-                if '\n' not in string[i+1:new_offset]:
-                    return None
-                return offset, new_offset, string[offset+1:i]
+                return offset, i+1, string[offset+1:i]
             elif escaped:
                 escaped = False
         return None
@@ -890,18 +924,6 @@ class Footnote(BlockToken):
             if key not in root.footnotes:
                 root.footnotes[key] = dest, title
 
-    @staticmethod
-    def backtrack(lines, string, offset):
-        """
-        Called when we iterated over some lines and found nothing
-        relevant on them. This returns those lines back to the parsing process.
-        """
-
-        # call lstrip() to prevent returning too many lines back (hence infinite loop), like in this case:
-        # * valid footlink definition line:      `[key]: valueN\r\n` (here `offset` points to `\r` after parsing the definition)
-        # * follow-up line containing just text: `something\n` (only this line should be re-processed and parsed as a Paragraph)
-        lines._index -= string[offset:].lstrip().count('\n')
-
 
 class ThematicBreak(BlockToken):
     """
@@ -930,7 +952,7 @@ class HTMLBlock(BlockToken):
         content (str): the raw HTML content.
     """
     _end_cond = None
-    multiblock = re.compile(r'<(script|pre|style)[ >\n]')
+    multiblock = re.compile(r'<(pre|script|style|textarea)[ >\n]')
     predefined = re.compile(r'<\/?(.+?)(?:\/?>|[ \n])')
     custom_tag = re.compile(r'(?:' + '|'.join((span_token._open_tag,
                                 span_token._closing_tag)) + r')\s*$')
@@ -943,7 +965,7 @@ class HTMLBlock(BlockToken):
         stripped = line.lstrip()
         if len(line) - len(stripped) >= 4:
             return False
-        # rule 1: <pre>, <script> or <style> tags, allow newlines in block
+        # rule 1: HTML tags designed to contain literal content, allow newlines in block
         match_obj = cls.multiblock.match(stripped)
         if match_obj is not None:
             cls._end_cond = '</{}>'.format(match_obj.group(1).casefold())
